@@ -232,8 +232,9 @@ struct backuppc_dirent {
 	unsigned char md5[16];
 	const char *name;
 	char fullpathname[BPC_MAXPATH_LEN];
+	char *symlinkdest;
+	char *hrdlinkdest;
 	size_t size;
-	ino_t inode;
  	dev_t dev;
 	mode_t mode;
 #ifndef _USE_WIN32_
@@ -241,7 +242,6 @@ struct backuppc_dirent {
 	uid_t uid;
 	gid_t gid;
 #else
-	uint32_t nlink;
 	uint32_t uid;
 	uint32_t gid;
 #endif
@@ -255,10 +255,18 @@ struct backuppc_dirent {
 	int fd;
 };
 
+struct backuppc_inode_list;
+struct backuppc_inode_list {
+	ino_t inode_num;
+	char *fullpathname;
+	struct backuppc_inode_list *_next;
+};
+
 /*
  * Private structure handle used by backuppc_opendir() and backuppc_closedir()
  */
 typedef struct backuppc_dirhandle {
+	struct backuppc_inode_list *ilist[4097];
 	struct backuppc_dirent ret;
 	char *pat_exclude;
 	char *pat_include;
@@ -268,6 +276,8 @@ typedef struct backuppc_dirhandle {
 	int isfile;
 } BPC_DIR;
 
+
+
 char *sha1sum(char *string) {
 	unsigned char digest[20];
 	static char ret[(sizeof(digest) * 2) + 1] = {0};
@@ -407,6 +417,7 @@ static BPC_DIR *backuppc_opendir(char *pathname) {
 	BPC_DIR *dh;
 	struct stat stbuf;
 	int stat_ret;
+	int i;
 
 	if (pathname[0] != '/') {
 		return(NULL);
@@ -422,6 +433,10 @@ static BPC_DIR *backuppc_opendir(char *pathname) {
 	dh->pat_include = NULL;
 	dh->isfile = 0;
 
+	for (i = 0; i < (sizeof(dh->ilist) / sizeof(dh->ilist[0])); i++) {
+		dh->ilist[i] = NULL;
+	}
+
 #ifdef _USE_WIN32_
 	if (strlen(pathname) < 2) {
 		dh->pathname = strdup("/");
@@ -462,6 +477,9 @@ static BPC_DIR *backuppc_opendir(char *pathname) {
 	dh->lastdrivecheck = 0;
 	dh->pathname = strdup(pathname);
 
+	dh->ret.symlinkdest = NULL;
+	dh->ret.hrdlinkdest = NULL;
+
 	return(dh);
 }
 
@@ -489,6 +507,7 @@ static BPC_DIR *backuppc_opendir(char *pathname) {
  *
  */
 static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
+	struct backuppc_inode_list *cur_inode;
 	struct dirent *dent;
 	struct stat stbuf;
 	size_t pathnamelen;
@@ -499,7 +518,8 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 	char drive_letter_str[32];
 	int i, drive_exists;
 #endif
-	int stat_ret;
+	int stat_ret, readlink_ret;
+	int hash_idx;
 
 	if (!dh) {
 		return(NULL);
@@ -535,7 +555,7 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 				dh->ret.name = strdup(drive_letter_str);
 				dh->ret.type = BPC_DIRENT_DIR;
 				dh->ret.size = 0;
-				dh->ret.inode = ((0xf << ((sizeof(dh->ret.inode) * 8) - 4)) | ((1 << ((sizeof(dh->ret.inode) * 8) - 1)) - 1)) - i;
+//				dh->ret.inode = ((0xf << ((sizeof(dh->ret.inode) * 8) - 4)) | ((1 << ((sizeof(dh->ret.inode) * 8) - 1)) - 1)) - i;
 				dh->ret.md5_set = 0;
 				dh->ret.fd = -1;
 				dh->ret.dev = 0;
@@ -556,6 +576,10 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 
 	pathnamelen = strlen(dh->pathname);
 
+	/*
+	 * A zero-length pathname never makes sense, and it will break some
+	 * calculations below.
+	 */
 	if (pathnamelen < 1) {
 		return(NULL);
 	}
@@ -605,11 +629,9 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 		if (!dh->ret.name) {
 			continue;
 		}
-		dh->ret.inode = stbuf.st_ino;
 		dh->ret.dev = stbuf.st_dev;
 		dh->ret.rdev = stbuf.st_rdev;
 		dh->ret.mode = stbuf.st_mode;
-		dh->ret.nlink = stbuf.st_nlink;
 		dh->ret.uid = stbuf.st_uid;
 		dh->ret.gid = stbuf.st_gid;
 #ifdef _USE_WIN32_
@@ -631,6 +653,15 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 #ifdef S_IFLNK
 			case S_IFLNK:
 				dh->ret.type = BPC_DIRENT_SYMLINK;
+#ifdef HAVE_READLINK
+				if (!dh->ret.symlinkdest) {
+					dh->ret.symlinkdest = malloc(BPC_MAXPATH_LEN);
+				}
+				readlink_ret = readlink(dh->ret.fullpathname, dh->ret.symlinkdest, BPC_MAXPATH_LEN);
+				if (readlink_ret >= 0 && readlink_ret < BPC_MAXPATH_LEN) {
+					dh->ret.symlinkdest[readlink_ret] = '\0';
+				}
+#endif
 				break;
 #endif
 			case S_IFCHR:
@@ -655,6 +686,31 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
 				break;
 		}
 
+		/*
+		 * If a file has a link count greater than one, determine if
+		 * we have seen it before, if not note that we've seen it now
+		 * so we can look for it later.
+		 */
+		if (stbuf.st_nlink > 1 && dh->ret.type != BPC_DIRENT_DIR && dh->ret.type != BPC_DIRENT_UNKNOWN) {
+			hash_idx = stbuf.st_ino % (sizeof(dh->ilist) / sizeof(dh->ilist[0]));
+			for (cur_inode = dh->ilist[hash_idx]; cur_inode; cur_inode = cur_inode->_next) {
+				if (cur_inode->inode_num == stbuf.st_ino) {
+					dh->ret.type = BPC_DIRENT_HRDLINK;
+					dh->ret.hrdlinkdest = cur_inode->fullpathname;
+					break;
+				}
+			}
+
+			if (dh->ret.type != BPC_DIRENT_HRDLINK) {
+				cur_inode = malloc(sizeof(*cur_inode));
+				cur_inode->inode_num = stbuf.st_ino;
+				cur_inode->fullpathname = strdup(dh->ret.fullpathname);
+				cur_inode->_next = dh->ilist[hash_idx];
+				dh->ilist[hash_idx] = cur_inode;
+			}
+		}
+
+
 		if (dh->ret.type == BPC_DIRENT_FILE) {
 			dh->ret.size = stbuf.st_size;
 		} else {
@@ -688,6 +744,8 @@ static struct backuppc_dirent *backuppc_readdir(BPC_DIR *dh) {
  *
  */
 static void backuppc_closedir(BPC_DIR *dh) {
+	struct backuppc_inode_list *cur_inode, *next_inode;
+	int hash_idx;
 
 	if (!dh) {
 		return;
@@ -709,6 +767,22 @@ static void backuppc_closedir(BPC_DIR *dh) {
 		free(dh->pat_include);
 	}
 
+	for (hash_idx = 0; hash_idx < (sizeof(dh->ilist) / sizeof(dh->ilist[0])); hash_idx++) {
+		cur_inode = dh->ilist[hash_idx];
+		while (cur_inode) {
+			if (cur_inode->fullpathname) {
+				free(cur_inode->fullpathname);
+			}
+
+			next_inode = cur_inode->_next;
+
+			free(cur_inode);
+
+			cur_inode = next_inode;
+		}
+	}
+
+
 	free(dh);
 
 	return;
@@ -941,26 +1015,35 @@ static ssize_t backuppc_get_attr(struct backuppc_dirent *dent, char *outbuf, siz
 		buf += sizeof(modeval);
 	}
 
-#ifdef HAVE_READLINK
-	if (dent->type == BPC_DIRENT_SYMLINK) {
-		id = htons(BPC_ATTRID_SYMLINKDEST);
+	if (dent->type == BPC_DIRENT_SYMLINK && dent->symlinkdest) {
+		if ((outbuflen - (buf - outbuf)) >= (sizeof(id) + sizeof(len) + strlen(dent->symlinkdest))) {
+			id = htons(BPC_ATTRID_SYMLINKDEST);
+			memcpy(buf, &id, sizeof(id));
+			buf += sizeof(id);
 
-		memcpy(buf, &id, sizeof(id));
-		buf += sizeof(id);
-
-		len = readlink(dent->fullpathname, buf + sizeof(len), outbuflen - (buf - outbuf));
-		if (len < 1) {
-			buf -= sizeof(id);
-		} else {
-			len = htonl(len);
+			len = htonl(strlen(dent->symlinkdest));
 			memcpy(buf, &len, sizeof(len));
 			buf += sizeof(len);
 
-			len = ntohl(len);
-			buf += len;
+			memcpy(buf, dent->symlinkdest, strlen(dent->symlinkdest));
+			buf += strlen(dent->symlinkdest);
+		}
+	}
+
+	if (dent->type == BPC_DIRENT_HRDLINK && dent->hrdlinkdest) {
+		if ((outbuflen - (buf - outbuf)) >= (sizeof(id) + sizeof(len) + strlen(dent->hrdlinkdest))) {
+			id = htons(BPC_ATTRID_HRDLINKDEST);
+			memcpy(buf, &id, sizeof(id));
+			buf += sizeof(id);
+
+			len = htonl(strlen(dent->hrdlinkdest));
+			memcpy(buf, &len, sizeof(len));
+			buf += sizeof(len);
+
+			memcpy(buf, dent->hrdlinkdest, strlen(dent->hrdlinkdest));
+			buf += strlen(dent->hrdlinkdest);
 		}
 	}
-#endif
 
 	/* If we can't fit the termination marker in, scrap the entire buffer. */
 	if ((outbuflen - (buf - outbuf)) >= sizeof(id)) {
